/* Copyright (C) 1999 Donald J Bindner
 * Copyright (C) 2004 Martin Renold
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/* I would like to acknowledge Jouni.Lohikoski@iki.fi whose utility
 * tleds influenced some of the look and feel for this program and
 * whose source code served as my example for process-handling.
 * (by Donald)
 */

/* I just added POP3/IMAP support by stealing pop.* and socklib.*
 * from asmail-0.56. It was quiet easy, thanks to MoGul
 * (by Martin)
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#include <signal.h>
#include <errno.h>

#include <unistd.h>
#include <dirent.h>

#include <pthread.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include "pop.h" /* stolen from the asmail project */

#define VERSION "0.8"
#define MYNAME "mailblink"
#define PIDFILE "/var/run/mailblink.pid"
#define LOCKFILE "/var/run/.mailblink-lock"

#define WARN 0
#define MESSAGE 1

static const char *light[11] = { "num", "caps", "scroll", "parport0", "parport1", "parport2", "parport3", "parport4", "parport5", "parport6", "parport7" };

/* These control blinking in another thread */
extern int lednum;
extern int blinks;
extern void *led_thread( void *not_used );
/* proto for the check loop */
void check_loop( void );

struct mbox_ll_t {
    struct mbox_ll_t* next;
    const char * filename;
    time_t old_mtime;
    int blinks;
};

typedef struct mbox_ll_t mbox_ll;

/* files to watch */
static mbox_ll* mboxes = NULL;

/* pop3 mailbox to poll */
static char *pop3setupfile = NULL;
static char pop3string[255];
/* delay between checks */
static int delay = -1;
#define DEFAULT_DELAY_LOCAL 30
#define DEFAULT_DELAY_POP 300

/* file descriptor used to lock pid-file */
static int fd;
/* pids for each light */
static pid_t pid[11] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

/* saved euid */
static uid_t saved_euid;

/* some user options */
static int opt_k=0, opt_D;

/* default verbosity */
int verbosity = 1;

void help_exit( void ) {
    printf( "%s version %s\t\tDon Bindner <don@bindner.com>\n\
Usage: %s [options]\n\
\t-d <delay>\tDelay between checks (range: 10 to 3600 seconds)\n\
\t-f <file/dir>\tMail file or directory to watch.\n\
\t-p <file>\tFile describing POP3/IMAP account to watch\n\
\t         \tit must contain one single line in this form:\n\
\t         \t  ::popserver username passwd [port]\n\
\t         \t  >>imapserver username passwd [port] [mailbox]\n\
\t-k\t\tKill old processes.\n\
\t-l <led>\tLed to blink (num, caps or scroll, parport0 - parport7)\n\
\t-q\t\tDecrease verbosity.  Repeat for even less.\n\
\t-v\t\tIncrease verbosity.  Repeat for even more.\n\
\t-D\t\tDebug.  Don't fork into the background.\n", MYNAME, VERSION, MYNAME );
    exit( 1 );
}

void message( int level, const char *format, ... ) {
    va_list ap;

    if( verbosity < level ) return;

    va_start( ap, format );
    vfprintf( stdout, format, ap );
    fflush( stdout );
}

int add_mbox( const char* newmboxname ) {
    mbox_ll *newmbox;

    /* Make a new head for the list */
    newmbox = (mbox_ll*)malloc(sizeof(mbox_ll));
    if (newmbox == NULL) {
	return 1;
    }

    /* Save the mbox */
    newmbox->filename = newmboxname;
    newmbox->blinks = 0;
    newmbox->old_mtime = time(NULL);

    /* Add the old list as the tail of the new list */
    newmbox->next = mboxes;
    mboxes = newmbox;

    return 0;
}

void cleanup_mboxes() {
    mbox_ll *ptr;

    while (mboxes) {
	ptr = mboxes;
	mboxes = mboxes->next;
	free(ptr);
    }
}

int pid_file_exists( void ) {
    struct stat status;

    message( 2, "Checking for existing PID file.\n" );
    return !( stat( PIDFILE, &status ));
}

int lock_pid_file( void ) {
    message( 2, "Locking PID file.\n" );
    return fd = open( LOCKFILE, O_WRONLY|O_CREAT|O_EXCL|O_SYNC, 0666);
}

void unlock_pid_file( void ) {
    message( 2, "Unlocking PID file.\n" );
    assert( fd != -1 );
    if( close( fd )) assert(0);
    if( unlink( LOCKFILE )) assert( 0 );
}

void read_pid_file( void ) {
    char str[255];
    int i;
    FILE *f;

    message( 2, "Reading PID file.\n" );

    if( (f = fopen( PIDFILE, "r" )) != NULL) {
	for( i = 0; i < 11; i++ ) {
	    fgets( str, sizeof(str), f );
	    pid[i] = strtol( str, NULL, 10 );
	}
	fclose( f );
    }
}

void write_pid_file( void ) {
    int i;
    FILE *f;

    message( 2, "Writing PID file.\n" );

    if( (f = fopen( PIDFILE, "w" ) ) != 0) {
	for( i = 0; i < 11; i++ ) {
	    fprintf( f, "%d %s\n", pid[i], light[i] );
	}
	fclose( f );
    }
}

void kill_old_process( int lednum ) {
/*
 * Will kill the (old) process who's PID is read from a
 * valid PID file.
 */

    message( 2, "Killing old process: %d.\n", pid[lednum] );

    /* Check first for process to kill */
    if( !pid[lednum] ) {
	message( MESSAGE, "No process to kill.\n" );
	return;
    }

    /* lower privilege */
    seteuid( getuid());

    if( kill( pid[lednum], SIGTERM )) {
	switch( errno ) {
	    case ESRCH:		/* process was already gone */
		pid[lednum] = 0;
		break;
	    case EPERM:
		message( WARN, "Permission denied killing PID %d.\n", pid[lednum] );
		break;
	    default:
		assert( 0 );
	}
    } else {
	message( MESSAGE, "Process %ld killed.\n", pid[lednum] );
	pid[lednum] = 0;
    }

    /* recover privilege */
    seteuid( saved_euid );
}

int old_process_alive( int lednum ) {
    char procFileName[256];
    struct stat status;

    message( 2, "Checking process status for PID: %d.\n", pid[lednum] );

    sprintf( procFileName, "/proc/%ld/environ", (long)pid[lednum] );
    return( !stat(procFileName, &status ));
}

void remove_pid() {
    message( 2, "Removing PID from PID file: %d.\n", pid[lednum] );

    if( lock_pid_file() == -1 ) exit( 1 );

    read_pid_file();
    pid[lednum] = 0;
    write_pid_file();
    unlock_pid_file();
}

void my_exit( void ) {
    blinks = 0;

    message( 2, "my_exit() called\n" );

    /* recover privilege */
    seteuid( saved_euid );

    cleanup_mboxes();

    remove_pid();
}

int semid;

/* raise the semaphore */
int up( void ) {
    struct sembuf sbuf;

    sbuf.sem_num = 0;
    sbuf.sem_op = 1;
    sbuf.sem_flg = 0;

    return semop( semid, &sbuf, 1 );
}

/* lower the appropriate semaphore */
int down( void ) {
    struct sembuf sbuf;

    sbuf.sem_num = 0;
    sbuf.sem_op = -1;
    sbuf.sem_flg = 0;

    return semop( semid, &sbuf, 1 );
}

void my_handler( int sig ) {
    static int done=0;
    int flag;

    /* only let this handler be called from one thread? */
    down();
    flag = done++;
    up();
    if( flag ) return;
    
    message( 2, "my_handler() called\n" );
    exit( 1 );
}

int main( int argc, char *argv[] ) {
    int c, i;
    pid_t childpid;
    sigset_t sigset;
    pthread_t thread;
    char *defmailbox;
    mbox_ll *ptr;

    /* So we can drop privilege temporarily */
    saved_euid = geteuid();
    seteuid( saved_euid );
    
    lednum = 2; /* default to blink scroll lock */

    /* block signals from delivery until we are ready to handle them */
    if( sigemptyset( &sigset )) assert(0);
    if( sigaddset( &sigset, SIGTERM )) assert(0);
    if( sigaddset( &sigset, SIGINT )) assert(0);
    if( sigaddset( &sigset, SIGHUP )) assert(0);
    if( sigaddset( &sigset, SIGQUIT )) assert(0);
    if( sigaddset( &sigset, SIGTTOU )) assert(0);
    if( sigprocmask( SIG_BLOCK, &sigset, NULL )) assert(0);

    /* Get all of the command line switches */
    while( EOF != ( c = getopt( argc, argv, "Dd:f:hkl:p:qv" ))) {
	switch( c ) {
	    case 'D':
		opt_D = 1;
		break;
	    case 'd':
		delay = strtol( optarg, NULL, 10 );
		break;
	    case 'f':
		if (add_mbox(optarg)) {
		    message( WARN, "Can't add mailbox to list\n" );
		    exit(1);
		}
		break;
            case 'p':
      	        pop3setupfile = optarg;
		break;
	    case 'h':
		help_exit();
		break;
	    case 'k':
		opt_k = 1;
		break;
	    case 'l':
		lednum = -1;
		for( i = 0; i < 11; i++ ) {
		    if( !strcmp( optarg, light[i] )) lednum = i;
		}
		if( lednum == -1 ) help_exit();
		break;
	    case 'q':
		verbosity--;	/* lower verbosity */
		break;
	    case 'v':
		verbosity++;	/* increase verbosity */
		break;
	    default:
		help_exit();
	}
    }

    /* Might as well get this out of the way */
    if( geteuid()) {
	message( WARN, "This program requires root privelege to run.\n" );
	exit( 1 );
    }

    /* use POP3/IMAP */
    if (pop3setupfile) {
      FILE *f;
    
      message( 2, "Reading POP3/IMAP setup file.\n" );
    
      /* drop privilege */
      seteuid( getuid());

      if( (f = fopen( pop3setupfile, "r" )) != 0 ) {
	/* FIXME: reading user data -- buffer overflow? I guess not. */
	fgets( pop3string, sizeof(pop3string), f );
	fclose( f );
	if ( strlen(pop3string) <= 3 ) {
	  message( WARN, "Bad POP3/IMAP setup file %s\n", optarg );
	  message( WARN, "The file should contain one single line in this form:\n" );
	  message( WARN, "::popserver username passwd [port]\n" );
	  message( WARN, "or, if using imap:\n" );
	  message( WARN, ">>imapserver username passwd [port] [mailbox]\n" );
	  message( WARN, "(port and mailbox are optional)\n" );	  
	  exit(1);
	}

        message( 2, "String read from config file: %s\n", pop3string );
      } else {
	message( WARN, "Could not read POP3/IMAP setup file %s", optarg );
	/* recover privilege (perhaps not necessary)*/
	seteuid( saved_euid );
	exit(1);
      }

      /* recover privilege */
      seteuid( saved_euid );
    }

    /* Let's make sure that delay has reasonable defaults */
    if( delay == -1 ) {
      delay = pop3setupfile ? DEFAULT_DELAY_POP : DEFAULT_DELAY_LOCAL;
    }
    if( delay > 3600 ) delay = 3600;
    if( delay < 10 ) delay = 10;
    message( MESSAGE, "Delay between polls set to %d seconds.\n", delay );

    /* Lock pid file and read it -- DON'T EXIT UNTIL THIS LOCK IS REMOVED */
    if( lock_pid_file() == -1 ) {
	message( WARN, "Could not create lockfile: %s\n", LOCKFILE );
	exit( 1 );
    }
    read_pid_file();

    /* kill an existing process */
    if( opt_k ) {
	kill_old_process( lednum );
	write_pid_file();
	unlock_pid_file();
	exit( 0 );
    }

    /* someone may be already using this led */
    if( pid[lednum] ) {
	if( old_process_alive( lednum )) {
	    message( WARN, "Led %s already used by PID: %d\n", 
		    light[lednum], pid[lednum] );
	    unlock_pid_file();
	    exit( 1 );
	} else {
	    /* process was dead */
	    message( 2, "Old pid was already dead: %d\n", pid[lednum] );
	    pid[lednum] = 0;
	    write_pid_file();
	}
    }

    if (mboxes == NULL && pop3setupfile == NULL) {
	/* We'll start with the file from $MAIL but if the user didn't 
	   overide */
	defmailbox = getenv( "MAIL" );
	if (defmailbox != NULL)
	    if (add_mbox( defmailbox )) {
		message( WARN, "Can't add default mailbox to list\n" );
		exit(1);
	    }
    }

    if( mboxes == NULL && pop3setupfile == NULL) {
	message( WARN, "Don't know what files to monitor.  Specify a file with -f.\n" );
	unlock_pid_file();
	exit( 1 );
    } else {
      if ( pop3setupfile ) {
	message( MESSAGE, "Preparing to monitor a POP3/IMAP Mailbox on %s LED.\n", light[lednum] );
      } else {
	for(ptr = mboxes; ptr != NULL; ptr = ptr->next) {
	    message( MESSAGE, "Preparing to monitor %s on %s LED.\n", 
		ptr->filename, light[lednum] );
	}
      }
    }

    if( opt_D ) {	/* Run in foreground to debug */
	message( MESSAGE, "Debug mode--running in foreground.\n" );
	childpid = getpid();
    } else {	/* fork into the background */
	if( (childpid = fork()) == -1 ) {
	    message( WARN, "%s: fork() failed", MYNAME );

	    unlock_pid_file();
	    exit( 1 );
	}
    }

    if( childpid ) {	/* parent code */
	/* update the pid file */
	pid[lednum] = childpid;
	write_pid_file();
	unlock_pid_file();

	message( MESSAGE, "%s running with PID %ld.\n", MYNAME, childpid );
	if( !opt_D ) exit( 0 );		/* only exit if not debugging */
    }

    /* detatch ourself from a controlling terminal and mounted filesystems */
    if( !opt_D ) {	/* if not debugging */
	setsid();
	chdir( "/" );
	umask( 0 );
    }

    /* create thread to do the blinking */
    pthread_create( &thread, NULL, led_thread, NULL );

    /* drop privilege */
    seteuid( getuid());
    
    /* Set up our exit protocol */
    if( atexit(my_exit)) assert(0);

    /* Set up a semaphore for exit handler */
    semid = semget( IPC_PRIVATE, 1, IPC_CREAT | IPC_EXCL | 0600 );
    if( semid == -1 ) {
	message( WARN, "Unable to get semaphore.\n" );
    }
    semctl( semid, 0, SETVAL, 1 );
    
    /* set signal handler and deliver all blocked signals */
    {
	struct sigaction act;

	memset( &act, 0, sizeof( struct sigaction ));
	act.sa_handler = my_handler;
	act.sa_flags |= SA_RESTART;
	sigfillset( &act.sa_mask );
	if( sigaction( SIGTERM, &act, NULL )) assert(0);
	if( sigaction( SIGINT, &act, NULL )) assert(0);
	if( sigaction( SIGHUP, &act, NULL )) assert(0);
	if( sigaction( SIGQUIT, &act, NULL )) assert(0);
	if( sigaction( SIGTTOU, &act, NULL )) assert(0);
	if( sigprocmask( SIG_UNBLOCK, &sigset, NULL )) assert(0);
    }

    /* continue indefinitely */
    check_loop();

    exit( 0 );
}


void check_loop( void ) {
    struct stat stat_buf;
    int i;
    const char *filename;
    mbox_ll *ptr;

    message( 3, "check_loop() started.\n" );

    while( 1 ) {
      if (pop3setupfile) {
	/* not used - just there because I didn't want to change pop.c */
	int old, new;
	message( MESSAGE, "Polling POP / IMAP mailbox.\n");
	/* pop3string (extracted from the asmail manpage)
               [ ::popserver username passwd <port> ]
               [ >>imapserver username passwd <port> <mailbox> ]
	*/
	blinks = pop(pop3string, &old, &new);
	message( MESSAGE, "pop() reported: mailbox got %d new mails.\n", blinks);
      } else {
	/* Mailfile / Maildir part */
	for ( ptr = mboxes; ptr != NULL; ptr = ptr->next ) {
	    filename = ptr->filename;
	
	    message( 5, "Opening %s.\n", filename );
	    if( stat( filename, &stat_buf )) {	/* file does not exist */
	      message( 5, "%s does not exist.\n", filename );
	      ptr->blinks = 0;
	    } else { 				/* file exists */
	      if( S_ISDIR( stat_buf.st_mode )) {	/* maildir support */
		DIR *d;
		
		d = opendir( filename );	/* open directory */
		if( d ) {
		  message( 5, "Directory %s opened.\n", filename );
		  i = -2;	/* don't count . and .. */
		  while ( readdir( d )) i++;	/* count files */
		  closedir( d );
		  
		  ptr->blinks = i;
		} else {
		  message( 5, "Could not open directory %s.\n", filename );
		  ptr->blinks = 0;
		}
	      } else if( stat_buf.st_atime < stat_buf.st_mtime ) { /* been modified */
		if( !ptr->blinks || ( stat_buf.st_mtime > ptr->old_mtime )) {
		  message( 5, " modified.\n" );
		  ptr->blinks++;
		  ptr->old_mtime = stat_buf.st_mtime;
		}
	      } else {		/* has been read */
		message( 5, " read.\n" );
		ptr->blinks = 0;
	      }
	    }

	}
	i=0;
	for ( ptr = mboxes; ptr != NULL; ptr = ptr->next )
	    i += ptr->blinks;
	
	blinks = i;
      }
      sleep( delay );
    }
}

